"""Dummy simulation intended for testing.
"""

from warnings import warn

import numpy as np
import pandas as pd



class SimulationDummy:
    """Dummy for Simulation testing.
    Only for testing if the program works, do not expect realistic behavior!"""
    def __init__(
        self,
        reference:pd.Series,
        param_spec:pd.DataFrame,
        result_spec:pd.DataFrame,
        seed=1234,
    ) -> None:
        self.param_spec = param_spec
        if not {
            "ResultFile",
            "ResultReach",
            "ResultNode",
            "ResultColumn",
            "Timestep",
            "Start",
            "End",
        }.issubset(result_spec.columns):
            raise ValueError(
                "result_spec needs columns: ResultFile,ResultReach,ResultNode,ResultColumn,Timestep,Start,End"
            )
        if not (
            result_spec["Start"].dt.tz is None and result_spec["End"].dt.tz is None
        ):
            raise ValueError(
                "Timezone for Start and End has to be None. Use dt.tz_localize(None) to drop timezone."
            )
        self.reference = reference
        rng = np.random.default_rng(seed=seed)
        # these parameters present ideal values, where the reference is returned
        self.ideal_params = pd.Series(index=param_spec.index,data=rng.uniform(param_spec["Min"],param_spec["Max"]))
        # these values 
        self.deviations = pd.DataFrame(index=reference.index,columns=param_spec.index,data=0.0)
        for p in param_spec.index:
            for l in reference.index.get_level_values("Label").unique():
                sn = slice(None)
                ref_part = reference.loc[(l,sn)]
                amp = rng.uniform(0,ref_part.max())
                center = rng.uniform(0,1)
                x_stretch = rng.uniform(1,20)
                self.deviations.loc[(l,sn),p]=_make_bell_array(len(ref_part),center,amp,x_stretch)
    
    def close(self):
        pass


    # the ** means that simulation run may be called like so:
    # result = simulation_run(key1=p1,key2=p2, ...)
    # but kwparams will contain a dict of form {"key1":p1,"key2":p2,...}
    # if you called the simulation like simulation_run(key1=p1,key2=p2, ...),
    # key1, key2, ... would need to be valid python names
    # but with simulation_run(**{"key1":p1,"key2":p2,...}) key1, key2 could be any string.
    # Because the optimizer uses the second form, we can use the index from param_specs as a key,
    # even though the index may not be a valid python name.
    def simulation_run(self, **kwparams):
        """ Dummy simulation.
        Takes parameters as arguments and returns the 
        dummy results as a series with MultiIndex[label, time].
        Note that kwparams needs keys that match the index of DataFrame self.param_specs."""

        if not  set(kwparams.keys()) == set(self.param_spec.index):
            raise ValueError("parameters need to match self.param_specs.")
        # squared normalized difference from ideal parameters is used to multiply the deviation
        dev_factor = ((pd.Series(kwparams)-self.ideal_params)/(self.param_spec["Max"]-self.param_spec["Min"]))**2
        return (self.deviations*dev_factor).sum(axis="columns")+self.reference.copy()


    def read_results(self) -> pd.Series:
        """Reads all results according to event table. Reads each file at most once
        returns:
            a pandas Series with (label,Datetime) as index.
        """
        return self.simulation_run(**self.param_spec["Default"])


def _make_bell_array(length:int, center:float, amp:float, x_stretch:float):
    """Makes a numpy array which represents a bell curve"""
    lin_01 = np.linspace(0, 1, num=length)
    return amp / (1+((lin_01[..., np.newaxis] - center) * x_stretch) ** 2)